(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (global = global || self, factory(global.effector = {}));
}(this, function (exports) { 'use strict';

  /* eslint-disable */

  var invariant = function invariant(condition) {
    if (!condition) {
      var error = new Error('Minified exception occurred; use the non-minified dev environment ' + 'for the full error message and additional helpful warnings.');
      error.framesToPop = 1; // we don't care about invariant's own frame

      throw error;
    }
  };

  var invariant$1 = invariant;

  function symbolObservablePonyfill(root) {
  	var result;
  	var Symbol = root.Symbol;

  	if (typeof Symbol === 'function') {
  		if (Symbol.observable) {
  			result = Symbol.observable;
  		} else {
  			result = Symbol('observable');
  			Symbol.observable = result;
  		}
  	} else {
  		result = '@@observable';
  	}

  	return result;
  }

  /* global window */

  var root;

  if (typeof self !== 'undefined') {
    root = self;
  } else if (typeof window !== 'undefined') {
    root = window;
  } else if (typeof global !== 'undefined') {
    root = global;
  } else if (typeof module !== 'undefined') {
    root = module;
  } else {
    root = Function('return this')();
  }

  var result = symbolObservablePonyfill(root);

  //prettier-ignore
  const Kind = {
    store: 'store',
    event: 'event',
    effect: 'effect'
  };

  const stringRefcount = () => {
    let id = 0;
    return () => (++id).toString(36);
  };

  const nextID = stringRefcount();
  const Step = typeDef('step', {
    single: null,
    multi: null,
    seq: null,
    query: null
  });
  const Cmd = typeDef('cmd', {
    compute: null,
    emit: null,
    run: null,
    filter: null,
    update: null
  });
  const Ctx = typeDef('ctx', {
    compute: null,
    emit: null,
    run: null,
    filter: null,
    update: null
  }); //eslint-disable-next-line no-unused-vars

  function typeDef(group, t) {
    const result = {};

    for (const key in t) {
      result[key] = type.bind({
        key,
        group
      });
    }

    return result;
  }

  function type(data) {
    return {
      id: nextID(),
      type: this.key,
      group: this.group,
      data
    };
  }

  let id = 0;
  function createStateRef(current) {
    return {
      id: (++id).toString(36),
      current
    };
  }

  /* eslint-disable no-unused-vars */
  /* Step */

  function fx(tag, props) {
    if (tag in Cmd) {
      const tag_ = tag;
      return Step.single(Cmd[tag_](props));
    }

    const tag_ = tag;

    for (var _len = arguments.length, childrens = new Array(_len > 2 ? _len - 2 : 0), _key = 2; _key < _len; _key++) {
      childrens[_key - 2] = arguments[_key];
    }

    switch (tag_) {
      case 'seq':
        return Step.seq(childrens);

      case 'multi':
        return Step.multi(childrens);

      case 'single':
        return Step.single(childrens[0]);

      case 'query':
        return Step.query(props);
    } // if (typeof tag === 'function') return tag(props, childrens)
    // console.error('unknown node "%s"', tag)


    return null;
  }

  const newMeta = ctx => ({
    callstack: [],
    transactions: [],
    stop: false,
    ctx,
    reg: {
      isChanged: true
    },
    val: stateRefs({
      watch(ctx) {// console.log('watch', ctx)
      },

      ctx: null
    }),
    arg: null
  });

  const stateRefs = obj => {
    const result = {};

    for (const key in obj) {
      result[key] = createStateRef(obj[key]);
    }

    return result;
  };

  const emit = {
    cmd: meta => Ctx.emit({
      __stepArg: meta.arg
    }),
    transition: () => true
  };
  const filter = {
    cmd(meta) {
      const ctx = Ctx.filter({
        __stepArg: meta.arg
      });
      const runCtx = tryRun({
        err: false,
        result: null,
        arg: meta.arg,
        val: meta.val,
        fn: meta.callstack[meta.callstack.length - 1].data.data.filter
      });
      meta.reg.isChanged = Boolean(runCtx.result);
      return ctx;
    },

    transition: meta => Boolean(meta.reg.isChanged)
  };
  const run = {
    cmd(meta) {
      const ctx = Ctx.run({});
      if ('transactionContext' in meta.callstack[meta.callstack.length - 1].data.data) meta.transactions.push(meta.callstack[meta.callstack.length - 1].data.data.transactionContext(meta.arg));
      tryRun({
        err: false,
        result: null,
        arg: meta.arg,
        val: meta.val,
        fn: meta.callstack[meta.callstack.length - 1].data.data.runner
      });
      return ctx;
    },

    transition: () => false
  };
  const update = {
    cmd(meta) {
      let store;

      if ('val' in meta.callstack[meta.callstack.length - 1].data.data) {
        store = meta.val[meta.callstack[meta.callstack.length - 1].data.data.val] = meta.val[meta.callstack[meta.callstack.length - 1].data.data.val] || createStateRef(null);
      } else {
        store = meta.callstack[meta.callstack.length - 1].data.data.store;
      }

      store.current = meta.arg;
      return Ctx.update({
        __stepArg: meta.arg
      });
    },

    transition: () => true
  };
  const compute = {
    cmd(meta) {
      const newCtx = Ctx.compute({
        __stepArg: null
      });
      const runCtx = tryRun({
        err: false,
        result: null,
        arg: meta.arg,
        val: meta.val,
        fn: meta.callstack[meta.callstack.length - 1].data.data.fn
      });
      meta.reg.isChanged = !runCtx.err;

      if (!runCtx.err) {
        newCtx.data.__stepArg = runCtx.result;
      }

      return newCtx;
    },

    transition: meta => Boolean(meta.reg.isChanged)
  };

  const tryRun = ctx => {
    try {
      ctx.result = ctx.fn.call(null, ctx.arg, ctx.val);
    } catch (err) {
      console.error(err);
      ctx.err = true;
    }

    return ctx;
  };

  var command = /*#__PURE__*/Object.freeze({
    emit: emit,
    filter: filter,
    run: run,
    update: update,
    compute: compute
  });

  const querySome = (ctx, meta) => {
    const data = meta.callstack[meta.callstack.length - 1].data;
    const fn = data.fn;
    const arg = meta.arg;
    const queryResult = fn(arg, ctx, meta);
    const newCtx = Ctx.emit({
      __stepArg: queryResult.arg
    });
    const items = queryResult.list;
    meta.arg = queryResult.arg;

    for (let i = 0; i < items.length; i++) {
      runStep(items[i], newCtx, meta);
    }

    meta.arg = arg;
    meta.stop = false;
  };

  const queryShape = (ctx, meta) => {
    const data = meta.callstack[meta.callstack.length - 1].data;
    const fn = data.fn;
    const shape = data.shape;
    const arg = meta.arg;
    const queryResult = fn(arg, ctx, meta);

    for (const key in queryResult) {
      if (!(key in shape)) continue;
      const def = shape[key];
      const subArg = queryResult[key];
      const newCtx = Ctx.emit({
        __stepArg: subArg
      });
      meta.arg = subArg;
      runStep(def, newCtx, meta);
    }

    meta.arg = arg;
    meta.stop = false;
  };

  const single = (ctx, meta) => {
    meta.val.ctx.current = ctx;
    meta.arg = ctx.data.__stepArg;
    meta.ctx = command[meta.callstack[meta.callstack.length - 1].data.type].cmd(meta);
    meta.stop = !command[meta.ctx.type].transition(meta);
  };
  const multi = (ctx, meta) => {
    const items = meta.callstack[meta.callstack.length - 1].data.slice();

    for (let i = 0; i < items.length; i++) {
      runStep(items[i], ctx, meta);
    }

    meta.stop = false;
    return;
  };
  const seq = (ctx, meta) => {

    meta.ctx = ctx;
    const items = meta.callstack[meta.callstack.length - 1].data.slice();

    for (let i = 0; i < items.length; i++) {
      runStep(items[i], meta.ctx, meta);

      if (meta.stop) {
        break;
      }
    }
  };
  const query = (ctx, meta) => {
    switch (meta.callstack[meta.callstack.length - 1].data.mode) {
      case 'some':
        return querySome(ctx, meta);

      case 'shape':
        return queryShape(ctx, meta);
    }
  }; //extendWalk(runStep, stepVisitor)

  var stepVisitor = /*#__PURE__*/Object.freeze({
    single: single,
    multi: multi,
    seq: seq,
    query: query
  });

  function runStep(step, ctx, meta) {
    meta.stop = false;
    meta.callstack.push(step);
    stepVisitor[step.type](ctx, meta);
    meta.callstack.pop();
  }

  function walkEvent(payload, event) {
    walkNode(event.graphite.seq, Ctx.emit({
      __stepArg: payload
    }));
  }
  function walkNode(seq, ctx) {
    const meta = newMeta(ctx); // __val('active', 'run')

    runStep(seq, meta.ctx, meta);
    runTransactions(meta.transactions);
  }

  const runTransactions = transactions => {
    for (let i = 0; i < transactions.length; i++) {
      (0, transactions[i])();
    }

    transactions.length = 0;
  };

  function Name(shortName, fullName, path) {
    this.shortName = shortName;
    this.fullName = fullName;
    this.path = path;
  }

  function createName(name, parent) {
    let path;
    let fullName;
    const shortName = name;

    if (parent === undefined) {
      if (name.length === 0) {
        path = [];
      } else {
        path = [name];
      }

      fullName = name;
    } else {
      if (name.length === 0) {
        path = parent.path;
        fullName = parent.fullName;
      } else {
        path = parent.path.concat([name]);

        if (parent.fullName.length === 0) {
          fullName = name;
        } else {
          fullName = '' + parent.fullName + '/' + name;
        }
      }
    }

    return new Name(shortName, fullName, path);
  }

  const noopIndexOf = () => -1; //eslint-disable-next-line no-unused-vars


  const noopSplice = (i, n) => [];

  function disposer() {
    const i = this.indexOf();
    if (i === -1) return;
    this.splice(i, 1);
    this.indexOf = noopIndexOf;
    this.splice = noopSplice;
  }

  const createWatcher = opts => {
    const subscribers = opts.parent.next.data;
    const instance = {
      indexOf: subscribers.indexOf.bind(subscribers, opts.child),
      splice: subscribers.splice.bind(subscribers)
    };
    const result = disposer.bind(instance);
    result.unsubscribe = disposer.bind(instance);
    return result;
  };

  function forward(opts) {
    const toSeq = opts.to.graphite.seq;
    const fromGraphite = opts.from.graphite;
    fromGraphite.next.data.push(toSeq);
    return createWatcher({
      child: toSeq,
      parent: fromGraphite
    });
  }

  //@jsx fx
  const nextID$1 = stringRefcount();
  function eventFabric(_ref) {
    let nameRaw = _ref.name,
        parent = _ref.parent;
    const id = nextID$1();
    const name = nameRaw || id;
    const fullName = makeName(name, parent);
    const compositeName = createName(name, parent);
    const graphite = fabric({
      fullName
    }); //$off

    const instance = function instance(payload) {
      for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
        args[_key - 1] = arguments[_key];
      }

      return instance.create(payload, fullName, args);
    };

    instance.getType = () => compositeName.fullName //eslint-disable-next-line no-unused-vars
    ;

    instance.create = (payload, fullName, args) => {
      walkEvent(payload, instance);
      return payload;
    };

    instance.kind = Kind.event;

    instance[result] = () => instance;

    instance.id = id;
    instance.watch = watchEvent.bind(null, instance);
    instance.map = mapEvent.bind(null, instance);
    instance.filter = filterEvent.bind(null, instance);
    instance.prepend = prepend.bind(null, instance);
    instance.subscribe = subscribe.bind(null, instance);
    instance.graphite = graphite;
    instance.shortName = name;
    instance.domainName = parent;
    instance.compositeName = compositeName;
    return instance;
  }

  function subscribe(event, observer) {
    return event.watch(payload => observer.next(payload));
  }

  function prepend(event, fn) {
    const contramapped = eventFabric({
      name: '* → ' + event.shortName,
      parent: event.domainName
    });
    forward({
      from: contramapped,
      to: {
        graphite: {
          seq: fx("seq", null, fx("compute", {
            fn: function (_fn) {
              function fn(_x) {
                return _fn.apply(this, arguments);
              }

              fn.toString = function () {
                return _fn.toString();
              };

              return fn;
            }(newValue => fn(newValue))
          }), event.graphite.seq)
        }
      }
    });
    return contramapped;
  }

  function mapEvent(event, fn) {
    const mapped = eventFabric({
      name: '' + event.shortName + ' → *',
      parent: event.domainName
    });
    forward({
      from: event,
      to: {
        graphite: {
          seq: fx("seq", null, fx("compute", {
            fn: function (_fn2) {
              function fn(_x2) {
                return _fn2.apply(this, arguments);
              }

              fn.toString = function () {
                return _fn2.toString();
              };

              return fn;
            }(newValue => fn(newValue))
          }), mapped.graphite.seq)
        }
      }
    });
    return mapped;
  }

  function filterEvent(event, fn) {
    const mapped = eventFabric({
      name: '' + event.shortName + ' →? *',
      parent: event.domainName
    });
    forward({
      from: event,
      to: {
        graphite: {
          seq: fx("seq", null, fx("compute", {
            fn: function (_fn3) {
              function fn(_x3) {
                return _fn3.apply(this, arguments);
              }

              fn.toString = function () {
                return _fn3.toString();
              };

              return fn;
            }(newValue => fn(newValue))
          }), fx("filter", {
            filter: result$$1 => result$$1 !== undefined
          }), mapped.graphite.seq)
        }
      }
    });
    return mapped;
  }

  function watchEvent(event, watcher) {
    return forward({
      from: event,
      to: {
        graphite: {
          seq: fx("run", {
            runner: newValue => watcher(newValue, event.getType())
          })
        }
      }
    });
  }

  function makeName(name, compositeName) {
    const fullName = compositeName === null || compositeName === void 0 ? void 0 : compositeName.fullName;

    if (!fullName) {
      if (!name) return '';
      return name;
    }

    return '' + fullName + '/' + name;
  }

  const fabric = args => {
    const nextSteps = fx("multi", null);
    return {
      next: nextSteps,
      seq: fx("seq", null, fx("emit", {
        fullName: args.fullName
      }), nextSteps)
    };
  };

  //@jsx fx
  function relayShape(opts) {
    const query = opts.query;
    const map = opts.shape;
    const from = opts.from;
    const shape = {};

    for (const key in map) {
      shape[key] = map[key].graphite.seq;
    }

    const unsub = forward({
      from,
      to: {
        graphite: {
          seq: fx("seq", null, fx("query", {
            mode: "shape",
            shape: shape,
            fn: query
          }))
        }
      }
    });
    return unsub;
  } //eslint-disable-next-line no-unused-vars

  function relay(opts, queryFn) {
    if (typeof queryFn === 'function') return relayRaw({
      from: opts,
      query: queryFn
    });
    return relayRaw(opts);
  }

  function relayRaw(opts) {
    const query = opts.query;
    const from = opts.from;
    const fn = queryFn.bind(query);
    const unsub = forward({
      from,
      to: {
        graphite: {
          seq: fx("seq", null, fx("query", {
            mode: "some",
            fn: fn
          }))
        }
      }
    });
    return unsub;

    function queryFn(arg, ctx, meta) {
      const query = this;
      const queryResult = query(arg);
      const queryList = queryResult.list;
      const resultList = [];

      for (let i = 0; i < queryList.length; i++) {
        const item = queryList[i];
        if (Boolean(item)) resultList.push(item.graphite.seq);
      }

      const result = {};
      result.arg = queryResult.arg;
      result.list = resultList;
      return result;
    }
  }

  function createEvent(name) {
    return eventFabric({
      name
    });
  }

  function fromObservable(observable) {
    invariant$1(typeof observable === 'function' || typeof observable === 'object' && observable !== null, 'expect observable to be object or function');
    invariant$1( //$off
    result in observable, 'expect observable to have property Symbol.observable'); //$off

    const observableItem = observable[result]();
    const event = createEvent();
    observableItem({
      next: event
    });
    return event;
  }

  function withProps(store, handler) {
    return props => handler(store.getState(), props);
  }
  function getDisplayName(store) {
    if (store.compositeName) {
      return store.compositeName.fullName;
    }

    if (store.domainName) {
      return store.domainName.fullName;
    }

    return store.id;
  }

  const supportsUserTiming = typeof performance !== 'undefined' && typeof performance.mark === 'function' && typeof performance.clearMarks === 'function' && typeof performance.measure === 'function' && typeof performance.clearMeasures === 'function';

  //@jsx fx
  function reset(storeInstance, event) {
    return on.call(this, storeInstance, event, () => storeInstance.defaultState);
  }
  function getState(storeInstance) {
    return storeInstance.plainState.current;
  }
  function off(storeInstance, event) {
    const currentSubscription = storeInstance.subscribers.get(event);
    if (currentSubscription === undefined) return;
    currentSubscription();
    storeInstance.subscribers.delete(event);
  }

  const readName = e => {
    switch (e.kind) {
      case Kind.store:
        return e.shortName;

      default:
        return e.getType();
    }
  };

  function on(storeInstance, event, handler) {
    const e = event;
    storeInstance.subscribers.set(e, forward({
      from: e,
      to: {
        graphite: {
          seq: fx("seq", null, fx("compute", {
            fn: newValue => {
              const lastState = getState(storeInstance);
              return handler(lastState, newValue, readName(e));
            }
          }), fx("filter", {
            filter: data => {
              const lastState = getState(storeInstance);
              return data !== lastState && data !== undefined;
            }
          }), storeInstance.graphite.seq)
        }
      }
    }));
    return this;
  }
  function observable(storeInstance) {
    const result$$1 = {
      subscribe(observer) {
        invariant$1(typeof observer === 'object' && observer !== null, 'Expected the observer to be an object.');

        function observeState(state) {
          if (observer.next) {
            observer.next(state);
          }
        }

        return subscribe$1(storeInstance, observeState);
      }

    }; //$off

    result$$1[result] = function () {
      return this;
    };

    return result$$1;
  }
  function watch(storeInstance, eventOrFn, fn) {
    const kind = String((eventOrFn === null || eventOrFn === void 0 ? void 0 : eventOrFn.kind) || '__');

    switch (kind) {
      case 'store':
      case 'event':
      case 'effect':
        invariant$1(typeof fn === 'function', 'watch requires function handler');
        return eventOrFn.watch(payload => //$todo
        fn(getState(storeInstance), payload, readName(eventOrFn)));

      case '__':
      default:
        invariant$1(typeof eventOrFn === 'function', 'watch requires function handler');
        return subscribe$1(storeInstance, eventOrFn);
    }
  }
  function subscribe$1(storeInstance, listener) {
    invariant$1(typeof listener === 'function', 'Expected the listener to be a function');
    let lastCall = getState(storeInstance);

    try {
      listener(lastCall);
    } catch (err) {
      console.error(err);
    }
    return forward({
      from: storeInstance,
      to: {
        graphite: {
          seq: //$todo
          fx("run", {
            runner: args => {

              if (args === lastCall) {
                return;
              }

              lastCall = args;

              try {
                listener(args);
              } catch (err) {
                console.error(err);
              }
            }
          })
        }
      }
    });
  }
  function thru(fn) {
    return fn(this);
  }
  function dispatch(action) {
    return action;
  }
  function mapStore(store, _fn, firstState) {
    let lastValue = store.getState();
    let lastResult;

    try {
      lastResult = _fn(lastValue, firstState);
    } catch (err) {
      console.error(err);
    }
    const innerStore = this({
      name: '' + store.shortName + ' → *',
      currentState: lastResult,
      parent: store.domainName
    });
    forward({
      from: store,
      to: {
        graphite: {
          seq: fx("seq", null, fx("compute", {
            fn: newValue => {
              lastValue = newValue;
              const lastState = innerStore.getState();
              let result$$1;

              try {
                result$$1 = _fn(newValue, lastState);
              } catch (err) {
                console.error(err);
              }
              return result$$1;
            }
          }), fx("filter", {
            filter: result$$1 => {
              const lastState = innerStore.getState();
              const isChanged = result$$1 !== lastState && result$$1 !== undefined;

              if (isChanged) {
                lastResult = result$$1;
              }
              return isChanged;
            }
          }), innerStore.graphite.seq)
        }
      }
    });
    return innerStore;
  }

  //@jsx fx
  function storeFabric(props) {
    const currentState = props.currentState,
          name = props.name,
          parent = props.parent;
    const plainState = createStateRef(currentState);
    const currentId = plainState.id;
    const defaultState = currentState;
    const updater = createEvent('update ' + currentId);
    const storeInstance = {
      graphite: createDef(plainState),
      kind: Kind.store,
      id: currentId,
      shortName: currentId,
      domainName: parent,
      defaultState,
      plainState,
      subscribers: new Map(),
      compositeName: createName(name || '', parent)
    };
    const store = {
      compositeName: storeInstance.compositeName,
      graphite: storeInstance.graphite,
      kind: Kind.store,
      id: currentId,
      shortName: currentId,
      domainName: parent,
      setState,
      off: off.bind(null, storeInstance),
      watch: watch.bind(null, storeInstance),
      subscribe: subscribe$1.bind(null, storeInstance),
      getState: getState.bind(null, storeInstance)
    };
    store.reset = reset.bind(store, storeInstance);
    store.on = on.bind(store, storeInstance);
    store.defaultState = defaultState;
    store.map = mapStore.bind(storeFabric, store);
    store.thru = thru.bind(store);
    store.dispatch = dispatch.bind(null); //$off

    store[result] = observable.bind(null, storeInstance);
    store.on(updater, (_, payload) => payload);

    function setState(value, reduce) {
      const currentReducer = typeof reduce === 'function' ? reduce : (_, payload) => payload;
      const state = getState(storeInstance);
      const newResult = currentReducer(state, value);
      updater(newResult);
    }

    return store;
  }

  function filterBeforeUpdate(newValue) {
    return newValue !== undefined && newValue !== this.current;
  }

  const createDef = plainState => {
    const def = {};
    def.next = fx("multi", null);
    def.seq = fx("seq", null, fx("filter", {
      filter: filterBeforeUpdate.bind(plainState)
    }), fx("update", {
      store: plainState
    }), def.next);
    return def;
  };

  function createStore(state) {
    return storeFabric({
      currentState: state
    });
  }

  const storeObjectMaxNames = 25;
  function storeObjectArrayName(arr) {
    let i = 0;
    const max = storeObjectMaxNames - 1;
    const maxLength = arr.length - 1;
    let name = 'combine(';

    for (const store of arr) {
      const comma = i === max || maxLength === i ? '' : ', ';

      if (store.kind !== Kind.store) {
        name += store.toString() + comma;
      } else {
        name += getDisplayName(store) + comma;
      }

      i += 1;
      if (comma === '') break;
    }

    name += ')';
    return name;
  }
  function storeObjectName(obj) {
    let i = 0;
    const keys = Object.keys(obj);
    const max = storeObjectMaxNames - 1;
    const maxLength = keys.length - 1;
    let name = 'combine(';

    for (const key in obj) {
      const comma = i === max || maxLength === i ? '' : ', ';
      const store = obj[key];

      if (store.kind !== Kind.store) {
        name += store.toString() + comma;
      } else {
        name += getDisplayName(store) + comma;
      }

      i += 1;
      if (comma === '') break;
    }

    name += ')';
    return name;
  }
  function setStoreName(store, rawName) {
    const compositeName = createName(rawName, store.domainName); //TODO: fix this
    //store.shortName = rawName

    if (!store.compositeName) {
      store.compositeName = compositeName;
      return;
    }

    store.compositeName.path = compositeName.path;
    store.compositeName.shortName = compositeName.shortName;
    store.compositeName.fullName = compositeName.fullName;
  }

  //@jsx fx

  function createStoreArray(obj) {
    const state = [...obj];
    const stateNew = [...obj];
    const updater = createEvent('update ' + Math.random().toString());
    let updates = [];

    const committer = () => {
      updates = [];
      return () => {
        let current = store.getState();

        for (const fn of updates) {
          current = fn(current);
        }

        commit = committer();
        updater(current);
      };
    };

    let commit = committer();

    for (const _ref of state.map((e, i) => [i, e])) {
      const key = _ref[0];
      const child = _ref[1];

      if (child.kind === Kind.store) {
        const substore = child; //eslint-disable-next-line no-unused-vars

        const runCmd = fx("run", {
          runner: newValue => {}
        });

        runCmd.data.data.transactionContext = data => {
          updates.push(state => {
            const nextState = [...state];
            nextState[key] = data;
            return nextState;
          });
          return commit;
        };

        stateNew[key] = substore.getState(); //$todo

        forward({
          from: substore,
          to: {
            graphite: {
              seq: runCmd
            }
          }
        });
      }
    }

    const name = storeObjectArrayName(obj);
    const store = storeFabric({
      name,
      currentState: stateNew
    }); //$todo

    store.defaultShape = obj;
    store.on(updater, (_, payload) => payload);
    return store;
  }

  function createStoreObjectMap(obj) {
    const state = Object.assign({}, obj);
    const stateNew = Object.assign({}, obj);
    const updater = createEvent('update ' + Math.random().toString());
    let updates = [];

    const committer = () => {
      updates = [];
      return () => {
        let current = store.getState();

        for (const fn of updates) {
          current = fn(current);
        }

        commit = committer();
        updater(current);
      };
    };

    let commit = committer();

    for (const _ref2 of Object.entries(state)) {
      const key = _ref2[0];
      const child = _ref2[1];
      const substore = child;
      if (substore.kind !== Kind.store) continue; //eslint-disable-next-line no-unused-vars

      const runCmd = fx("run", {
        runner: newValue => {}
      });

      runCmd.data.data.transactionContext = data => {
        updates.push(state => Object.assign({}, state, {
          [key]: data
        }));
        return commit;
      };

      stateNew[key] = substore.getState(); //$todo

      forward({
        from: substore,
        to: {
          graphite: {
            seq: runCmd
          }
        }
      });
    } //$todo


    const name = storeObjectName(obj);
    const store = storeFabric({
      name,
      currentState: stateNew
    }); //$todo

    store.defaultShape = obj;
    store.on(updater, (_, payload) => payload);
    return store;
  }

  function createStoreObject(obj) {
    if (Array.isArray(obj)) {
      return createStoreArray(obj);
    }

    return createStoreObjectMap(obj);
  } //eslint-disable-next-line

  function extract(store, extractor) {
    let result;
    if ('defaultShape' in store) result = extractor(store.defaultShape);else result = extractor(store.defaultState);
    return createStoreObject(result);
  }

  function createApi(store, setters) {
    const result = {};

    for (const _ref of Object.entries(setters)) {
      const key = _ref[0];
      const handler = _ref[1];
      const event = createEvent(key);
      store.on(event, handler);
      result[key] = event;
    }

    return result;
  }

  function restoreObject(obj) {
    const result = {};

    for (const _ref of Object.entries(obj)) {
      const key = _ref[0];
      const value = _ref[1];
      result[key] = createStore(value);
    }

    return result;
  }
  function restoreEffect(event, defaultState) {
    const store = createStore(defaultState);
    store.on(event.done, (_, _ref2) => {
      let result = _ref2.result;
      return result;
    });
    return store;
  }
  function restoreEvent(event, defaultState) {
    const store = createStore(defaultState);
    store.on(event, (_, v) => v);
    return store;
  } //eslint-disable-next-line no-unused-vars

  const visitorRestore = {
    store: (_ref3) => {
      let obj = _ref3.obj;
      return obj;
    },
    event: (_ref4) => {
      let obj = _ref4.obj,
          defaultState = _ref4.defaultState;
      return restoreEvent(obj, defaultState);
    },
    effect: (_ref5) => {
      let obj = _ref5.obj,
          defaultState = _ref5.defaultState;
      return restoreEffect(obj, defaultState);
    },
    __: (_ref6) => {
      let obj = _ref6.obj;
      return restoreObject(obj);
    }
  };
  function restore(obj, defaultState) {
    return visitorRestore[String((obj === null || obj === void 0 ? void 0 : obj.kind) || '__')]({
      obj,
      defaultState
    });
  }

  function combine() {
    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    invariant$1(args.length > 0, 'at least one argument required');
    const handler = args[args.length - 1];
    const stores = args.slice(0, -1);
    const structStore = createStoreObject(stores);
    return structStore.map(list => handler(...list));
  }

  /* eslint-disable */
  /**
   * Similar to invariant but only logs a warning if the condition is not met.
   * This can be used to log issues in development environments in critical
   * paths. Removing the logging code for production environments will keep the
   * same logic and follow the same code paths.
   */

  var warning = function warning()
  /*::condition: any, format: string, args: any*/
  {};

  var warning$1 = warning;

  const exec = (args, cbs) => {
    let syncError;
    /*:: = throwSymbol*/

    let req;
    /*:: = doneSymbol*/

    let successSync = false;
    let fpromise;

    try {
      req = cbs[2](args);
      successSync = true;
    } catch (err) {
      syncError = err;
    }

    if (successSync === false) {
      fpromise = Promise.reject(syncError); //TODO deprecate this
      //$off

      fpromise.cache = () => undefined;

      const anyway = Promise.resolve(undefined); //$off

      fpromise.anyway = () => anyway;

      cbs[1]({
        params: args,
        error: syncError
      });
      return fpromise;
    }

    if (typeof req === 'object' && req !== null && typeof req.then === 'function') {
      const then = req;
      fpromise = then.then(result => {
        //$off
        fpromise.cache = () => result;

        cbs[0]({
          params: args,
          result
        });
        return result;
      }, error => {
        cbs[1]({
          params: args,
          error
        });
        throw error;
      });
      const anyway = fpromise.then(() => {}, () => {}); //$off

      fpromise.anyway = () => anyway;

      return fpromise;
    }

    const done = req;
    fpromise = Promise.resolve(done); //$off

    fpromise.cache = () => done;

    const anyway = Promise.resolve(undefined); //$off

    fpromise.anyway = () => anyway;

    cbs[0]({
      params: args,
      result: done
    });
    return fpromise;
  };

  const cbDone = function cbDone(value) {
    const cb = this[3];
    this.length = 0;
    cb(value);
  };

  const cbFail = function cbFail(value) {
    const cb = this[4];
    this.length = 0;
    cb(value);
  };

  const callbacks = (thunk, onDone, onFail) => [cbDone, cbFail, thunk, onDone, onFail];

  function effectFabric(_ref) {
    let name = _ref.name,
        domainName = _ref.domainName,
        parent = _ref.parent,
        handler = _ref.handler;
    //$off
    const instance = eventFabric({
      name,
      parent
    });
    const eventCreate = instance.create;
    const done = eventFabric({
      name: '' + instance.shortName + ' done',
      parent
    });
    const fail = eventFabric({
      name: '' + instance.shortName + ' fail',
      parent
    });
    instance.done = done;
    instance.fail = fail;

    instance.use = fn => {
      thunk = fn;
      return instance;
    };

    instance.use.getCurrent = () => thunk;

    instance.kind = Kind.effect;

    instance.create = (params, fullName, args) => {
      eventCreate(params, instance.getType(), args);
      return exec(params, callbacks(thunk, result => void done(result), error => void fail(error)));
    }; //eslint-disable-next-line no-unused-vars


    let thunk = handler || defaultThunk.bind(instance);
    return instance;
  } //eslint-disable-next-line no-unused-vars

  function defaultThunk(value) {
    warning$1(false, 'no thunk used in %s', this.getType());
    return Promise.resolve();
  }

  function createEffect(name, config) {
    const opts = {};
    opts.name = name;
    opts.domainName = '';
    opts.handler = config === null || config === void 0 ? void 0 : config.handler;
    return effectFabric(opts);
  }

  class DomainHistory {
    constructor() {
      this.events = new Set();
      this.effects = new Set();
      this.storages = new Set();
      this.domains = new Set();
    }

  }
  function domainHooks(history, compositeName, parentHooks) {
    let hooks;

    if (parentHooks) {
      hooks = childDomainHooks(parentHooks);
    } else {
      hooks = singleDomainHooks(compositeName);
    }

    hooks.domain.watch(domain => {
      history.domains.add(domain);
    });
    hooks.event.watch(event => {
      history.events.add(event);
    });
    hooks.store.watch(store => {
      history.storages.add(store);
    });
    hooks.effect.watch(effect => {
      history.effects.add(effect);
    });
    return hooks;
  }

  const singleDomainHooks = compositeName => {
    const event = eventFabric({
      name: `${compositeName.fullName} event hook`,
      parent: compositeName
    });
    const effect = eventFabric({
      name: `${compositeName.fullName} effect hook`,
      parent: compositeName
    });
    const store = eventFabric({
      name: `${compositeName.fullName} store hook`,
      parent: compositeName
    });
    const domain = eventFabric({
      parent: compositeName
    });
    return {
      event,
      effect,
      store,
      domain
    };
  };

  const childDomainHooks = parentHooks => {
    const event = parentHooks.event.prepend(_ => _);
    const effect = parentHooks.effect.prepend(_ => _);
    const store = parentHooks.store.prepend(_ => _);
    const domain = parentHooks.domain.prepend(_ => _);
    return {
      event,
      effect,
      store,
      domain
    };
  };

  const nextID$2 = stringRefcount();
  function domainFabric(nameRaw, parent, parentHooks) {
    const id = nextID$2();
    const name = nameRaw || '';
    const compositeName = createName(name, parent);
    const history = new DomainHistory();
    const hooks = domainHooks(history, compositeName, parentHooks);
    return {
      compositeName,
      id,

      getType() {
        return compositeName.fullName;
      },

      onCreateEvent(hook) {
        history.events.forEach(hook);
        return hooks.event.watch(hook);
      },

      onCreateEffect(hook) {
        history.effects.forEach(hook);
        return hooks.effect.watch(hook);
      },

      onCreateStore(hook) {
        history.storages.forEach(hook);
        return hooks.store.watch(hook);
      },

      onCreateDomain(hook) {
        history.domains.forEach(hook);
        return hooks.domain.watch(hook);
      },

      event(name) {
        const result = eventFabric({
          name,
          parent: compositeName
        });
        hooks.event(result);
        return result;
      },

      effect(name, config) {
        const result = effectFabric({
          name,
          domainName: compositeName.fullName,
          parent: compositeName,
          handler: config === null || config === void 0 ? void 0 : config.handler
        });
        hooks.effect(result);
        return result;
      },

      domain(name) {
        const result = domainFabric(name, compositeName, hooks);
        hooks.domain(result);
        return result;
      },

      store(state) {
        const result = storeFabric({
          currentState: state,
          parent: compositeName
        });
        hooks.store(result);
        return result;
      }

    };
  }

  function createDomain(name) {
    return domainFabric(name === undefined ? '' : name);
  }

  exports.combine = combine;
  exports.createDomain = createDomain;
  exports.createEvent = createEvent;
  exports.forward = forward;
  exports.fromObservable = fromObservable;
  exports.relay = relay;
  exports.relayShape = relayShape;
  exports.createEffect = createEffect;
  exports.createStore = createStore;
  exports.createStoreObject = createStoreObject;
  exports.setStoreName = setStoreName;
  exports.extract = extract;
  exports.createApi = createApi;
  exports.restore = restore;
  exports.restoreEvent = restoreEvent;
  exports.restoreEffect = restoreEffect;
  exports.restoreObject = restoreObject;
  exports.withProps = withProps;
  exports.Kind = Kind;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
//# sourceMappingURL=effector.umd.js.map
